/**
 * Copyright © 2014-2015 Paolo Simonetto
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package ocotillo.gui;

import ocotillo.geometry.Box;
import java.awt.Component;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;

/**
 * Listener that can be attached to a Component to implement Zoom and Pan
 * functionality. Based on code from Sorin Postelnicu.
 *
 */
public class ZoomAndPanListener implements MouseListener, MouseMotionListener, MouseWheelListener {

    private final Component targetComponent;
    private int currentZoomLevel = 0;
    private double zoomFactor = 1.2;
    private Point dragStartDeviceCoord;
    private AffineTransform transform = new AffineTransform();

    /**
     * Creates a zoom and pan listener.
     *
     * @param targetComponent the target component.
     */
    public ZoomAndPanListener(Component targetComponent) {
        this.targetComponent = targetComponent;
    }

    /**
     * Creates a zoom and pan listener with given zoom factor. The zoom factor
     * indicates the ratio dimension between elements at two consecutive zoom
     * levels.
     *
     * @param targetComponent the target component.
     * @param zoomFactor the zoom factor.
     */
    public ZoomAndPanListener(Component targetComponent, double zoomFactor) {
        this.targetComponent = targetComponent;
        this.zoomFactor = zoomFactor;
    }

    @Override
    public void mouseClicked(MouseEvent event) {
    }

    @Override
    public void mousePressed(MouseEvent event) {
        dragStartDeviceCoord = event.getPoint();
    }

    @Override
    public void mouseReleased(MouseEvent event) {
    }

    @Override
    public void mouseEntered(MouseEvent event) {
    }

    @Override
    public void mouseExited(MouseEvent event) {
    }

    @Override
    public void mouseMoved(MouseEvent event) {
    }

    @Override
    public void mouseDragged(MouseEvent event) {
        Point dragEndScreen = event.getPoint();
        pan(dragEndScreen);
    }

    @Override
    public void mouseWheelMoved(MouseWheelEvent event) {
        zoom(event.getPoint(), -event.getWheelRotation());
    }

    /**
     * Performs the pan operation.
     *
     * @param dragEndDeviceCoord the device coordinate at which the drag
     * movement ended.
     */
    private void pan(Point dragEndDeviceCoord) {
        Point2D dragStart = transformPoint(dragStartDeviceCoord);
        Point2D dragEnd = transformPoint(dragEndDeviceCoord);
        double dx = dragEnd.getX() - dragStart.getX();
        double dy = dragEnd.getY() - dragStart.getY();
        transform.translate(dx, dy);
        dragStartDeviceCoord = dragEndDeviceCoord;
        targetComponent.repaint();
    }

    /**
     * Performs the zoom operation.
     *
     * @param zoomCenterDeviceCoord the device coordinate at which the zoom
     * operation is applied.
     * @param zoomLevelIncrement the zoom level change. Positive numbers zoom
     * in, negative numbers zoom out.
     */
    private void zoom(Point zoomCenterDeviceCoord, int zoomLevelIncrement) {
        if (zoomLevelIncrement != 0) {
            currentZoomLevel = currentZoomLevel + zoomLevelIncrement;
            Point2D before = transformPoint(zoomCenterDeviceCoord);
            double factor = Math.pow(zoomFactor, zoomLevelIncrement);
            transform.scale(factor, factor);
            Point2D after = transformPoint(zoomCenterDeviceCoord);
            transform.translate(after.getX() - before.getX(), after.getY() - before.getY());
            targetComponent.repaint();
        }
    }

    /**
     * Transforms a point from device coordinates to user space ones.
     *
     * @param pointDeviceCoord the device coordinates of the point.
     * @return its user space coordinates.
     */
    private Point2D transformPoint(Point pointDeviceCoord) {
        try {
            AffineTransform inverse = transform.createInverse();
            Point2D result = new Point2D.Float();
            inverse.transform(pointDeviceCoord, result);
            return result;
        } catch (NoninvertibleTransformException ex) {
            throw new IllegalStateException("The transformation is not invertible");
        }
    }

    /**
     * Centers and zooms to fit the drawing nicely in the current component.
     *
     * @param boundingBox the bounding box of the drawing.
     */
    public void resetView(Box boundingBox) {
        transform = new AffineTransform();
        currentZoomLevel = 0;
        double xScaleRatio = targetComponent.getSize().width / boundingBox.width();
        double yScaleRatio = targetComponent.getSize().height / boundingBox.height();
        double scaleRatio = Math.min(xScaleRatio, yScaleRatio);
        int desiredZoomLevel = (int) Math.floor(Math.log(scaleRatio) / Math.log(zoomFactor));
        Point componentCenter = new Point(targetComponent.getWidth() / 2, targetComponent.getHeight() / 2);
        zoom(componentCenter, desiredZoomLevel);
        double dx = targetComponent.getWidth() / 2 - boundingBox.center().x();
        double dy = targetComponent.getHeight() / 2 + boundingBox.center().y();
        transform.translate(dx, dy);
    }

    /**
     * Gets the canvas transformation.
     *
     * @return the transformation.
     */
    public AffineTransform getTransform() {
        return transform;
    }

    /**
     * Sets the canvas transformation.
     *
     * @param coordTransform
     */
    public void setTransform(AffineTransform coordTransform) {
        this.transform = coordTransform;
    }

}
